cdef class UpdateRequest(BaseRequest):
    cdef:
        SchemaIndex index
        object key
        list operations


cdef char *encode_update_ops(WriteBuffer buffer,
                             char *p, list operations,
                             SchemaSpace space) except NULL
cdef int encode_request_update(WriteBuffer buffer,
                               SchemaSpace space, SchemaIndex index,
                               key_tuple, list operations,
                               bint is_upsert) except -1



import asyncio
import enum
import functools
import os
from typing import Optional, Type, Union

from .api import Api
from .exceptions import ErrorCode, TarantoolDatabaseError, TarantoolError
from .iproto import protocol
from .log import logger
from .stream import Stream

__all__ = ("Connection", "connect", "ConnectionState")


class ConnectionState(enum.IntEnum):
    CONNECTING = 1
    CONNECTED = 2
    RECONNECTING = 3
    DISCONNECTING = 4
    DISCONNECTED = 5


class Connection(Api):
    __slots__ = (
        "_host",
        "_port",
        "_username",
        "_password",
        "_fetch_schema",
        "_auto_refetch_schema",
        "_initial_read_buffer_size",
        "_encoding",
        "_connect_timeout",
        "_reconnect_timeout",
        "_request_timeout",
        "_ping_timeout",
        "_state",
        "_state_prev",
        "_transport",
        "_protocol",
        "_disconnect_waiter",
        "_reconnect_task",
        "_connect_lock",
        "_disconnect_lock",
        "_ping_task",
    )

    def __init__(
        self,
        *,
        host: str = "127.0.0.1",
        port: Union[int, str] = 3301,
        username: Optional[str] = None,
        password: Optional[str] = None,
        fetch_schema: bool = True,
        auto_refetch_schema: bool = True,
        connect_timeout: float = 3.0,
        request_timeout: float = -1.0,
        reconnect_timeout: float = 1.0 / 3.0,
        ping_timeout: float = 5.0,
        encoding: Optional[str] = None,
        initial_read_buffer_size: Optional[int] = None,
    ):
        """
        Connection constructor.

        To manipulate a Connection instance there are several functions:

            * await connect() - performs connecting, authorization and
                                schema fetching.

            * await disconnect() - performs disconnection.

            * close() - closes connection (not a coroutine)

        Connection also supports context manager protocol, which connects
        on entering and disconnecting on leaving a block.

        So one can simply use it as follows:

        .. code-block:: python

            async with asynctnt.Connection() as conn:
                await conn.call('box.info')

        :param host:
                Tarantool host (pass ``unix/`` to connect to unix socket)
        :param port:
                Tarantool port
                (pass ``/path/to/sockfile`` to connect ot unix socket)
        :param username:
                Username to use for auth
                (if ``None`` you are connected as a guest)
        :param password:
                Password to use for auth
        :param fetch_schema:
                Pass ``True`` to be able to use spaces and indexes names in
                data manipulation routines (default is ``True``)
        :param auto_refetch_schema:
                If set to ``True`` then when ER_WRONG_SCHEMA_VERSION error
                occurs on a request, schema is refetched and the initial
                request is resent. If set to ``False`` then schema will not
                be checked by Tarantool, so no errors will occur
        :param connect_timeout:
                Time in seconds how long to wait for connecting to socket
        :param request_timeout:
                Request timeout (in seconds) for all requests
                (by default there is no timeout)
        :param reconnect_timeout:
                Time in seconds to wait before automatic reconnect
                (set to ``0`` or ``None`` to disable auto reconnect)
        :param ping_timeout:
                If specified (default is 5 seconds) a background task
                will be created which will ping Tarantool instance
                periodically to check if it is alive and update schema
                if it is changed
                (set to ``0`` or ``None`` to disable this task)
        :param encoding:
                The encoding to use for all strings
                encoding and decoding (default is ``utf-8``)
        :param initial_read_buffer_size:
                Initial and minimum size of read buffer in bytes.
                Higher value means less reallocations, but higher
                memory usage (default is 131072).
        """
        super().__init__()
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._fetch_schema = False if fetch_schema is None else fetch_schema
        if auto_refetch_schema:  # None hack
            self._auto_refetch_schema = True
            if not self._fetch_schema:
                logger.warning(
                    "Setting fetch_schema to True as " "auto_refetch_schema is True"
                )
                self._fetch_schema = True
        else:
            self._auto_refetch_schema = False
        self._initial_read_buffer_size = initial_read_buffer_size
        self._encoding = encoding or "utf-8"

        self._connect_timeout = connect_timeout
        self._reconnect_timeout = reconnect_timeout or 0
        self._request_timeout = request_timeout
        self._ping_timeout = ping_timeout or 0

        self._transport = None
        self._protocol: Optional[protocol.Protocol] = None

        self._state = ConnectionState.DISCONNECTED
        self._state_prev = ConnectionState.DISCONNECTED
        self._disconnect_waiter = None
        self._reconnect_task = None
        self._connect_lock = asyncio.Lock()
        self._disconnect_lock = asyncio.Lock()
        self._ping_task = None

    def _set_state(self, new_state: ConnectionState):
        if self._state != new_state:
            logger.debug("Changing state %s -> %s", self._state.name, new_state.name)
            self._state_prev = self._state
            self._state = new_state

    def connection_lost(self, exc):
        if self._transport:
            self._transport.close()
        self._transport = None

        if self._disconnect_waiter:
            # disconnect() call happened
            self._disconnect_waiter.set_result(True)
            return

        # connection lost

        if self._reconnect_timeout > 0:
            # should reconnect
            self._start_reconnect(return_exceptions=False)
        else:
            # should not reconnect, close everything
            self.close()

    async def _ping_task_func(self):
        while self._state == ConnectionState.CONNECTED:
            try:
                await self.ping(timeout=2.0)
            except asyncio.CancelledError:
                break
            except Exception:
                pass

            await asyncio.sleep(self._ping_timeout)

    def _start_reconnect(self, return_exceptions: bool = False):
        if self._state in [ConnectionState.CONNECTING, ConnectionState.RECONNECTING]:
            logger.debug(
                "%s Cannot start reconnect: already reconnecting", self.fingerprint
            )
            return

        if self._reconnect_task:  # pragma: nocover
            return

        logger.info("%s Started reconnecting", self.fingerprint)
        self._set_state(ConnectionState.RECONNECTING)

        self._reconnect_task = asyncio.create_task(
            self._connect(return_exceptions=return_exceptions)
        )

    def protocol_factory(
        self,
        connected_fut: asyncio.Future,
        loop: asyncio.AbstractEventLoop,
        cls: Type[protocol.Protocol] = protocol.Protocol,
    ):
        return cls(
            host=self._host,
            port=self._port,
            username=self._username,
            password=self._password,
            fetch_schema=self._fetch_schema,
            auto_refetch_schema=self._auto_refetch_schema,
            request_timeout=self._request_timeout,
            initial_read_buffer_size=self._initial_read_buffer_size,
            encoding=self._encoding,
            connected_fut=connected_fut,
            on_connection_made=None,
            on_connection_lost=self.connection_lost,
            loop=loop,
        )

    async def _connect(self, return_exceptions: bool = True):
        loop = asyncio.get_running_loop()

        async with self._connect_lock:
            while True:
                try:
                    ignore_states = {
                        ConnectionState.CONNECTED,
                        ConnectionState.DISCONNECTING,  # disconnect() called
                    }

                    if self._state in ignore_states:
                        self._reconnect_task = None
                        return

                    self._set_state(ConnectionState.CONNECTING)

                    async def full_connect():
                        while True:
                            connected_fut = loop.create_future()

                            if self._host.startswith("unix/"):
                                unix_path = self._port
                                assert isinstance(unix_path, str), (
                                    "port must be a str instance for " "unix socket"
                                )
                                assert unix_path, "No unix file path specified"
                                assert os.path.exists(
                                    unix_path
                                ), "Unix socket `{}` not found".format(unix_path)

                                conn = loop.create_unix_connection(
                                    functools.partial(
                                        self.protocol_factory, connected_fut, loop
                                    ),
                                    unix_path,
                                )
                            else:
                                conn = loop.create_connection(
                                    functools.partial(
                                        self.protocol_factory, connected_fut, loop
                                    ),
                                    self._host,
                                    self._port,
                                )

                            tr, pr = await conn

                            try:
                                timeout = 0.05  # wait at least something
                                if self._connect_timeout is not None:
                                    timeout = self._connect_timeout / 2

                                await asyncio.wait_for(connected_fut, timeout=timeout)
                            except asyncio.TimeoutError:  # pragma: nocover
                                tr.close()
                                continue  # try again
                            except Exception:
                                tr.close()
                                raise

                            return tr, pr

                    tr, pr = await asyncio.wait_for(
                        full_connect(),
                        timeout=self._connect_timeout,
                    )

                    logger.info("%s Connected successfully", self.fingerprint)
                    self._set_state(ConnectionState.CONNECTED)

                    self._transport = tr
                    self._protocol = pr
                    self._set_db(self._protocol.get_common_db())
                    self._reconnect_task = None
                    self._normalize_api()

                    if self._ping_timeout:
                        self._ping_task = loop.create_task(self._ping_task_func())
                    return
                except TarantoolDatabaseError as e:
                    skip_errors = {
                        ErrorCode.ER_LOADING,
                        ErrorCode.ER_NO_SUCH_SPACE,
                        ErrorCode.ER_NO_SUCH_INDEX_ID,
                    }
                    if e.code in skip_errors:
                        # If Tarantool is still loading then reconnect
                        if self._reconnect_timeout > 0:
                            await self._wait_reconnect(e)
                            continue

                    if return_exceptions:
                        self._reconnect_task = None
                        raise e

                    logger.exception(e)
                    if self._reconnect_timeout > 0:
                        await self._wait_reconnect(e)
                        continue

                    return  # no reconnect, no return_exceptions
                except asyncio.CancelledError:
                    logger.debug("connect is cancelled")
                    self._reconnect_task = None
                    raise
                except Exception as e:
                    if self._reconnect_timeout > 0:
                        await self._wait_reconnect(e)
                        continue

                    if return_exceptions:
                        self._reconnect_task = None
                        raise e

                    logger.exception(e)
                    return  # no reconnect, no return_exceptions

    async def _wait_reconnect(self, exc: Optional[Exception] = None):
        self._set_state(ConnectionState.RECONNECTING)
        logger.warning(
            "Connect to %s failed: %s. Retrying in %f seconds",
            self.fingerprint,
            repr(exc) if exc else "",
            self._reconnect_timeout,
        )

        await asyncio.sleep(self._reconnect_timeout)

    async def connect(self) -> "Connection":
        """
        Connect coroutine
        """
        await self._connect(True)
        return self

    async def disconnect(self):
        """
        Disconnect coroutine
        """

        loop = asyncio.get_running_loop()

        async with self._disconnect_lock:
            if self._state == ConnectionState.DISCONNECTED:
                return

            self._set_state(ConnectionState.DISCONNECTING)

            logger.info("%s Disconnecting...", self.fingerprint)
            if self._reconnect_task:
                self._reconnect_task.cancel()
                self._reconnect_task = None

            if self._ping_task and not self._ping_task.done():
                self._ping_task.cancel()
                self._ping_task = None

            self._clear_db()
            if self._transport:
                self._disconnect_waiter = loop.create_future()
                self._transport.close()
                self._transport = None
                self._protocol = None

                await self._disconnect_waiter
                self._disconnect_waiter = None
                self._set_state(ConnectionState.DISCONNECTED)
            else:
                self._transport = None
                self._protocol = None
                self._disconnect_waiter = None
                self._set_state(ConnectionState.DISCONNECTED)

    def close(self):
        """
        Same as disconnect, but not a coroutine, i.e. it does not wait
        for disconnect to finish.
        """

        if self._state == ConnectionState.DISCONNECTED:
            return

        self._set_state(ConnectionState.DISCONNECTING)
        logger.info("%s Disconnecting...", self.fingerprint)

        if self._reconnect_task and not self._reconnect_task.done():
            self._reconnect_task.cancel()
            self._reconnect_task = None

        if self._ping_task and not self._ping_task.done():
            self._ping_task.cancel()
            self._ping_task = None

        if self._transport:
            self._transport.close()

        self._transport = None
        self._protocol = None
        self._disconnect_waiter = None
        self._clear_db()
        self._set_state(ConnectionState.DISCONNECTED)

    async def reconnect(self):
        """
        Reconnect coroutine.
        Just calls disconnect() and connect()
        """
        await self.disconnect()
        await self.connect()

    async def __aenter__(self):
        """
        Executed on entering the async with section.
        Connects to Tarantool instance.
        """
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Executed on leaving the async with section.
        Disconnects from Tarantool instance.
        """
        await self.disconnect()

    @property
    def fingerprint(self) -> str:
        return "Tarantool[{}:{}]".format(self._host, self._port)

    @property
    def host(self) -> str:
        """
        Tarantool host
        """
        return self._host

    @property
    def port(self) -> int:
        """
        Tarantool port
        """
        return self._port

    @property
    def username(self) -> Optional[str]:
        """
        Tarantool username
        """
        return self._username

    @property
    def password(self) -> Optional[str]:
        """
        Tarantool password
        """
        return self._password

    @property
    def fetch_schema(self) -> bool:
        """
        fetch_schema flag
        """
        return self._fetch_schema

    @property
    def auto_refetch_schema(self) -> bool:
        """
        auto_refetch_schema flag
        """
        return self._auto_refetch_schema

    @property
    def encoding(self) -> str:
        """
        Connection encoding
        """
        return self._encoding

    @property
    def reconnect_timeout(self) -> float:
        """
        Reconnect timeout value
        """
        return self._reconnect_timeout

    @property
    def connect_timeout(self) -> float:
        """
        Connect timeout value
        """
        return self._connect_timeout

    @property
    def request_timeout(self) -> float:
        """
        Request timeout value
        """
        return self._request_timeout

    @property
    def version(self) -> Optional[tuple]:
        """
        Protocol version tuple. ex.: (1, 6, 7)
        """
        if self._protocol is None:
            return None
        return self._protocol.get_version()

    @property
    def state(self) -> ConnectionState:
        """
        Current connection state

        :rtype: ConnectionState
        """
        return self._state

    @property
    def is_connected(self) -> bool:
        """
        Check if an underlying connection is active
        """
        if self._protocol is None:
            return False
        return self._protocol.is_connected()

    @property
    def is_fully_connected(self) -> bool:
        """
        Check if connection is fully active (performed auth
        and schema fetching)
        """
        if self._protocol is None:
            return False
        return self._protocol.is_fully_connected()

    @property
    def schema_id(self) -> Optional[int]:
        """
        Tarantool's current schema id
        """
        if self._protocol is None:
            return None
        return self._protocol.schema_id

    @property
    def schema(self) -> Optional[protocol.Schema]:
        """
        Current Tarantool schema with all spaces, indexes and fields
        """
        if self._protocol is None:  # pragma: nocover
            return None
        return self._protocol.schema

    @property
    def initial_read_buffer_size(self) -> int:
        """
        initial_read_buffer_size value
        """
        return self._initial_read_buffer_size

    async def refetch_schema(self):
        """
        Coroutine to force refetch schema
        """
        await self._protocol.refetch_schema()

    def _normalize_api(self):
        if (1, 6) <= self.version < (1, 7):  # pragma: nocover
            Api.call = Api.call16
            Connection.call = Connection.call16

        if self.version < (2, 10):  # pragma: nocover

            def stream_stub(_):
                raise TarantoolError("streams are available only in Tarantool 2.10+")

            Connection.stream = stream_stub

    def __repr__(self):
        return "<asynctnt.Connection host={} port={} state={}>".format(
            self.host, self.port, repr(self.state)
        )

    def stream(self) -> Stream:
        """
        Create new stream suitable for interactive transactions
        """
        stream = Stream()
        db = self._protocol.create_db(True)
        stream._set_db(db)
        return stream


async def connect(**kwargs) -> Connection:
    """
    connect shorthand. See :class:`asynctnt.Connection` for kwargs details

    :return: :class:`asynctnt.Connection` object
    """
    return await Connection(**kwargs).connect()



cdef enum ProtocolState:
    PROTOCOL_IDLE = 0
    PROTOCOL_GREETING = 1
    PROTOCOL_NORMAL = 2


cdef enum ConnectionState:
    CONNECTION_BAD = 0
    CONNECTION_CONNECTED = 1
    CONNECTION_FULL = 2


cdef class CoreProtocol:
    cdef:
        object host
        object port

        bytes encoding

        ProtocolState state
        ConnectionState con_state

        ReadBuffer rbuf
        tuple version
        bytes salt

    cdef bint _is_connected(self)
    cdef bint _is_fully_connected(self)

    cdef void _write(self, buf) except *
    cdef void _on_data_received(self, data)
    cdef void _process__greeting(self)
    cdef void _on_greeting_received(self)
    cdef void _on_response_received(self, const char *buf, uint32_t buf_len)
    cdef void _on_connection_made(self)
    cdef void _on_connection_lost(self, exc)



from libc.stdint cimport int8_t, int64_t, uint8_t, uint16_t, uint32_t, uint64_t
from libc.stdio cimport FILE


cdef extern from "../../third_party/msgpuck/msgpuck.h":
    cdef enum mp_type:
        MP_NIL = 0
        MP_UINT
        MP_INT
        MP_STR
        MP_BIN
        MP_ARRAY
        MP_MAP
        MP_BOOL
        MP_FLOAT
        MP_DOUBLE
        MP_EXT

    cdef uint8_t mp_load_u8(const char **data)
    cdef uint16_t mp_load_u16(const char **data)
    cdef uint32_t mp_load_u32(const char **data)
    cdef uint64_t mp_load_u64(const char **data)

    cdef char *mp_store_u8(char *data, uint8_t val)
    cdef char *mp_store_u16(char *data, uint16_t val)
    cdef char *mp_store_u32(char *data, uint32_t val)
    cdef char *mp_store_u64(char *data, uint64_t val)

    cdef mp_type mp_typeof(const char c)

    cdef uint32_t mp_sizeof_array(uint32_t size)
    cdef char *mp_encode_array(char *data, uint32_t size)
    cdef uint32_t mp_decode_array(const char **data)

    cdef uint32_t mp_sizeof_map(uint32_t size)
    cdef char *mp_encode_map(char *data, uint32_t size)
    cdef uint32_t mp_decode_map(const char **data)

    cdef uint32_t mp_sizeof_uint(uint64_t num)
    cdef char *mp_encode_uint(char *data, uint64_t num)
    cdef uint64_t mp_decode_uint(const char **data)

    cdef uint32_t mp_sizeof_int(int64_t num)
    cdef char *mp_encode_int(char *data, int64_t num)
    cdef int64_t mp_decode_int(const char **data)

    cdef uint32_t mp_sizeof_float(float num)
    cdef char *mp_encode_float(char *data, float num)
    cdef float mp_decode_float(const char **data)

    cdef uint32_t mp_sizeof_double(double num)
    cdef char *mp_encode_double(char *data, double num)
    cdef double mp_decode_double(const char **data)

    cdef char *mp_encode_strl(char *data, uint32_t len)
    cdef uint32_t mp_decode_strl(const char **data)

    cdef uint32_t mp_sizeof_str(uint32_t len)
    cdef char *mp_encode_str(char *data, const char *str, uint32_t len)
    cdef const char *mp_decode_str(const char **data, uint32_t *len)

    cdef char *mp_encode_binl(char *data, uint32_t len)
    cdef uint32_t mp_decode_binl(const char **data)

    cdef uint32_t mp_sizeof_bin(uint32_t len)
    cdef char *mp_encode_bin(char *data, const char *str, uint32_t len)
    cdef const char *mp_decode_bin(const char **data, uint32_t *len)

    cdef uint32_t mp_sizeof_extl(uint32_t len)
    cdef uint32_t mp_sizeof_ext(uint32_t len)
    cdef char *mp_encode_extl(char *data, int8_t type, uint32_t len)
    cdef char *mp_encode_ext(char *data, int8_t type, char *str, uint32_t len)
    cdef uint32_t mp_decode_extl(const char **data, int8_t *type)
    cdef const char *mp_decode_ext(const char **data, int8_t *type, uint32_t *len)

    cdef uint32_t mp_decode_strbinl(const char **data)
    cdef const char *mp_decode_strbin(const char **data, uint32_t *len)

    cdef char *mp_encode_nil(char *data)
    cdef void mp_decode_nil(const char **data)

    cdef char *mp_encode_bool(char *data, bint val)
    cdef bint mp_decode_bool(const char **data)

    cdef void mp_next(const char **data)

    cdef void mp_fprint(FILE *file, const char *data)



from .api import Api


class Stream(Api):
    def __init__(self):
        super().__init__()

    @property
    def stream_id(self) -> int:
        """
        Current stream is
        """
        return self._db.stream_id

    async def __aenter__(self):
        """
            If used as Context Manager `begin()` and `commit()`/`rollback()`
            are called automatically
        :return:
        """
        await self.begin()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Normally `commit()` is called on context manager exit, but
        in the case of exception `rollback()` is called
        """
        if exc_type and exc_val:
            await self.rollback()
        else:
            await self.commit()



import abc
import asyncio
import asyncio.subprocess
import functools
import logging
import math
import os
import random
import re
import select
import shutil
import socket
import string
import subprocess
import time
from threading import Thread
from typing import Optional

import yaml

__all__ = (
    "TarantoolInstanceProtocol",
    "TarantoolInstance",
    "TarantoolAsyncInstance",
    "TarantoolSyncInstance",
    "TarantoolSyncDockerInstance",
)

VERSION_STRING_REGEX = re.compile(r"\s*([\d.]+).*")


class TarantoolInstanceProtocol(asyncio.SubprocessProtocol):
    def __init__(self, tnt, on_exit):
        super().__init__()
        self._tnt = tnt
        self._on_exit = on_exit
        self._transport = None

    @property
    def logger(self):
        return self._tnt.logger

    @property
    def pid(self):
        return self._transport.get_pid() if self._transport else None

    def connection_made(self, transport):
        self.logger.info("Process started")
        self._transport = transport

    def pipe_data_received(self, fd, data):
        if not data:
            return
        line = data.decode()
        line = line.replace("\r", "")
        lines = line.split("\n")
        for line in lines:
            line = line.replace("\n", "")
            line = line.strip()
            if line:
                self.logger.info("=> %s", line)

    def process_exited(self):
        return_code = self._transport.get_returncode()
        if callable(self._on_exit):
            self._on_exit(return_code)

    @property
    def returncode(self):
        return self._transport.get_returncode()

    async def wait(self):
        """Wait until the process exit and return the process return code.

        This method is a coroutine."""
        return await self._transport._wait()

    def send_signal(self, signal):
        self._transport.send_signal(signal)

    def terminate(self):
        self._transport.terminate()

    def kill(self):
        self._transport.kill()


class TarantoolInstance(metaclass=abc.ABCMeta):
    def __init__(
        self,
        *,
        host="127.0.0.1",
        port=3301,
        console_host=None,
        console_port=3302,
        replication_source=None,
        title=None,
        logger=None,
        log_level=5,
        slab_alloc_arena=0.1,
        wal_mode="none",
        root=None,
        specify_work_dir=True,
        cleanup=True,
        initlua_template=None,
        applua="-- app.lua --",
        extra_box_cfg="",
        timeout=5.0,
        command_to_run="tarantool",
        command_args=None,
    ):
        """

        :param host: The host which Tarantool instance is going
                     to be listening on (default = 127.0.0.1)
        :param port: The port which Tarantool instance is going
                     to be listening on (default = 3301)
        :param console_host: The host which Tarantool console is going
                             to be listening on (to execute admin commands)
                             (default = host)
        :param console_port: The port which Tarantool console is going
                             to be listening on (to execute admin commands)
                             (default = 3302)
        :param replication_source: The replication source string.
                                   If it's None, then replication_source=nil
                                   in box.cfg
        :param title: Tarantool instance title (substitutes
                      into custom_proc_title). (default = "tnt[host:port]")
        :param logger: logger, where all messages are logged to.
                       default logger name = Tarantool[host:port]
        :param log_level: Tarantool's log_level (default = 5)
        :param slab_alloc_arena: Tarantool's slab_alloc_arena (default = 0.1)
        :param wal_mode: Tarantool's wal_mode (default = 'none')
        :param root: Tarantool's work_dir location
        :param specify_work_dir: Specify or not the workdir of Tarantool
        :param cleanup: do cleanup or not
        :param initlua_template: The initial init.lua template
                                 (default can be found in
                                 _create_initlua_template function)
        :param applua: Any extra lua script (a string)
                       (default = '-- app.lua --')
        :param extra_box_cfg: Any extra box.cfg params
                              (default = '')
        :param timeout: Timeout in seconds - how much to wait for tarantool
                        to become active
        :param command_to_run: command exe
        :param command_args: command args
        """

        self._host = host
        self._port = port
        self._console_host = console_host or host
        self._console_port = console_port
        self._replication_source = replication_source
        self._title = title or self._generate_title()
        self._logger = logger or logging.getLogger(self.fingerprint)
        self._log_level = log_level
        self._slab_alloc_arena = slab_alloc_arena
        self._wal_mode = wal_mode
        self._root = root or self._generate_root_folder_name()
        self._specify_work_dir = specify_work_dir
        self._cleanup = cleanup

        self._initlua_template = initlua_template or self._create_initlua_template()
        self._applua = applua
        self._extra_box_cfg = extra_box_cfg
        self._command_to_run = command_to_run
        self._command_args = command_args

        self._timeout = timeout
        self._is_running = False

    @property
    def replication_source(self):
        return self._replication_source

    @replication_source.setter
    def replication_source(self, value):
        self._replication_source = value

    def _random_string(
        self,
        length,
        *,
        source=string.ascii_uppercase + string.ascii_lowercase + string.digits,
    ):
        return "".join(random.choice(source) for _ in range(length))

    def _generate_title(self):
        return "tnt[{}:{}]".format(self._host, self._port)

    def _generate_root_folder_name(self):
        cwd = os.getcwd()
        path = None
        while path is None or os.path.isdir(path):
            folder_name = "__tnt__" + self._random_string(10)
            path = os.path.join(cwd, folder_name)
        return path

    @staticmethod
    def get_random_port():
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(("", 0))
        port = sock.getsockname()[1]
        sock.close()
        return port

    def _create_initlua_template(self):
        return """
            local function check_version(expected, version)
                -- from tarantool/queue compat.lua
                local fun = require 'fun'
                local iter, op  = fun.iter, fun.operator
                local function split(self, sep)
                    local sep, fields = sep or ":", {}
                    local pattern = string.format("([^%s]+)", sep)
                    self:gsub(pattern, function(c) table.insert(fields, c) end)
                    return fields
                end

                local function reducer(res, l, r)
                    if res ~= nil then
                        return res
                    end
                    if tonumber(l) == tonumber(r) then
                        return nil
                    end
                    return tonumber(l) > tonumber(r)
                end

                local function split_version(version_string)
                    local vtable  = split(version_string, '.')
                    local vtable2 = split(vtable[3],  '-')
                    vtable[3], vtable[4] = vtable2[1], vtable2[2]
                    return vtable
                end

                local function check_version_internal(expected, version)
                    version = version or _TARANTOOL
                    if type(version) == 'string' then
                        version = split_version(version)
                    end
                    local res = iter(version):zip(expected)
                                             :reduce(reducer, nil)
                    if res or res == nil then res = true end
                    return res
                end

                return check_version_internal(expected, version)
            end
            local cfg = {
              listen = "${host}:${port}",
              wal_mode = "${wal_mode}",
              custom_proc_title = "${custom_proc_title}",
              slab_alloc_arena = ${slab_alloc_arena},
              work_dir = ${work_dir},
              log_level = ${log_level},
              ${extra_box_cfg}
            }
            if check_version({1, 7}, _TARANTOOL) then
                cfg.replication = ${replication_source}
            else
                local repl = ${replication_source}
                if type(repl) == 'table' then
                    repl = table.concat(repl, ',')
                end
                cfg.replication_source = repl
            end
            require('console').listen("${console_host}:${console_port}")
            box.cfg(cfg)
            box.schema.user.grant("guest", "read,write,execute", "universe",
                                  nil, {if_not_exists = true})
            ${applua}
        """

    def _render_initlua(self):
        template = string.Template(self._initlua_template)
        if not self._replication_source:
            replication = "nil"
        elif isinstance(self._replication_source, str):
            replication = '"{}"'.format(self._replication_source)
        elif isinstance(self._replication_source, (list, tuple)):
            replication = ['"{}"'.format(e) for e in self._replication_source]
            replication = ",".join(replication)
            replication = "{" + replication + "}"
        else:
            raise TypeError("replication is of unsupported type")

        work_dir = "nil"
        if self._specify_work_dir:
            work_dir = '"' + self._root + '"'

        d = {
            "host": self._host,
            "port": self._port,
            "console_host": self._console_host,
            "console_port": self._console_port,
            "wal_mode": self._wal_mode,
            "custom_proc_title": self._title,
            "slab_alloc_arena": self._slab_alloc_arena,
            "replication_source": replication,
            "work_dir": work_dir,
            "log_level": self._log_level,
            "applua": self._applua if self._applua else "",
            "extra_box_cfg": self._extra_box_cfg,
        }
        return template.substitute(d)

    def _save_initlua(self, initlua):
        initlua = initlua.replace(" " * 4, "")
        initlua_path = os.path.join(self._root, "init.lua")
        with open(initlua_path, "w") as f:
            f.write(initlua)
        return initlua_path

    @property
    def logger(self):
        return self._logger

    @property
    def fingerprint(self):
        return "Tarantool[{}:{}]".format(self._host, self._port)

    def prepare(self, recreate):
        if recreate and os.path.exists(self._root):
            shutil.rmtree(self._root, ignore_errors=True)
        if not os.path.exists(self._root):
            os.mkdir(self._root)
        initlua = self._render_initlua()
        initlua_path = self._save_initlua(initlua)
        return initlua_path

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def console_port(self):
        return self._console_port

    @property
    def is_running(self):
        return self._is_running

    @property
    @abc.abstractmethod
    def pid(self):
        raise NotImplementedError

    @abc.abstractmethod
    def command(self, cmd, print_greeting=True):
        raise NotImplementedError

    @abc.abstractmethod
    def start(self, *, wait=True, recreate=True):
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self):
        raise NotImplementedError

    @abc.abstractmethod
    def terminate(self):
        raise NotImplementedError

    @abc.abstractmethod
    def kill(self):
        raise NotImplementedError

    def cleanup(self):
        self._is_running = False
        if self._cleanup:
            shutil.rmtree(self._root, ignore_errors=True)
        if self.host == "unix/":
            shutil.rmtree(self.port, ignore_errors=True)
        self._logger.info("Destroyed Tarantool instance (%s)", self._title)


class TcpSocket:
    BUFFER_SIZE = 1024

    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def connect(self):
        self._sock.connect((self._host, self._port))

    def close(self):
        if self._sock is not None:
            self._sock.close()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def write(self, data, flags=0):
        self._sock.sendall(data, flags)

    def read(self, n, flags=0):
        buf = bytearray()
        bytes_recd = 0

        while bytes_recd < n:
            chunk = self._sock.recv(self.BUFFER_SIZE, flags)
            if chunk == b"":
                raise RuntimeError("socket connection broken")
            buf.extend(chunk)

            bytes_recd += len(chunk)
        return bytes(buf)

    def read_until(self, separator=b"", flags=0):
        buf = bytearray()
        search_start = 0
        while True:
            chunk = self._sock.recv(self.BUFFER_SIZE, flags)
            if chunk == b"":
                raise RuntimeError("socket connection broken")
            buf.extend(chunk)
            pos = buf.find(separator, search_start)
            if pos != -1:
                return bytes(buf[: (pos + len(separator))])
            search_start = len(buf) - len(separator) - 1


class TarantoolSyncInstance(TarantoolInstance):
    WAIT_TIMEOUT = 5

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._process = None
        self._logger_thread = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    @property
    def pid(self):
        return self._process.pid if self._process is not None else None

    def start(self, *, wait=True, recreate=True):
        self._logger.info("Starting Tarantool instance (%s)", self._title)
        initlua_path = self.prepare(recreate)
        self._logger.info("Launching process")

        if not self._command_args:
            args = [self._command_to_run, initlua_path]
        else:
            args = [self._command_to_run, *self._command_args]

        flags = 0
        if os.name == "nt":
            flags |= subprocess.CREATE_NEW_PROCESS_GROUP
        self._process = subprocess.Popen(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, creationflags=flags
        )
        self._logger_thread = Thread(target=self._log_reader)
        self._logger_thread.start()

        if not wait:
            self._is_running = True
            return

        interval = 0.1
        attempts = math.ceil(self._timeout / interval)
        while attempts > 0:
            try:
                status = self.command("box.info.status", print_greeting=True)
                if status:
                    status = status[0]
                    if status == "running":
                        self._logger.info("Moved to the running state")
                        break
            except (OSError, RuntimeError):
                pass
            time.sleep(interval)
            attempts -= 1
        else:
            raise TimeoutError(
                "Timeout while waiting for Tarantool to move to running state"
            )
        self._is_running = True

    def _log_reader(self):
        def check_io():
            fds = []
            for h in [self._process.stdout, self._process.stderr]:
                if h is not None and not h.closed:
                    fds.append(h)
            if not fds:
                return False
            try:
                ready_to_read = select.select(fds, [], [], 10)[0]
            except (ValueError, OSError):
                # I/O operation on a closed socket
                return False
            for io in ready_to_read:
                if io.closed:
                    continue
                try:
                    line = io.readline()
                except ValueError:
                    # assuming it's just an fd error, so skip
                    # PyMemoryView_FromBuffer(): info->buf must not be NULL
                    continue
                line = line.decode()
                if len(line) > 0:
                    self._logger.info(line[:-1])
            return True

        while self._is_running and self._process.poll() is None:
            if not check_io():
                break
        check_io()

    def stop(self):
        if self._process is not None:
            self._process.terminate()
            self._logger.info("Waiting for process to complete")
            self._wait(self.WAIT_TIMEOUT, wait=True)
            self.cleanup()

    def terminate(self):
        if self._process is not None:
            self._process.terminate()
            self._wait(self.WAIT_TIMEOUT, wait=False)
            self.cleanup()

    def kill(self):
        if self._process is not None:
            self._process.kill()
            self.cleanup()

    def _wait(self, timeout, wait=True):
        if self._process:
            if wait:
                try:
                    self._process.wait(timeout)
                except subprocess.TimeoutExpired:
                    pass

            try:
                os.kill(self._process.pid, 0)
                self._process.kill()
                self.logger.warning("Force killed %s", self.fingerprint)
            except OSError:
                pass

    def cleanup(self):
        if self._process is not None:
            for h in [self._process.stdout, self._process.stderr, self._process.stdin]:
                if h is not None:
                    h.close()
        self._is_running = False
        if self._logger_thread is not None:
            self._logger_thread.join()
        super().cleanup()

    @staticmethod
    def _parse_version(version: str) -> Optional[tuple]:
        m = VERSION_STRING_REGEX.match(version)
        if m is not None:
            ver = m.group(1)
            return tuple(map(int, ver.split(".")))
        return None

    def version(self) -> Optional[tuple]:
        res = self.command("box.info.version")
        if not res:
            return None
        return self._parse_version(res[0])

    @property
    def bin_version(self) -> Optional[tuple]:
        proc = subprocess.Popen([self._command_to_run, "-V"], stdout=subprocess.PIPE)
        output = proc.stdout.read().decode()
        version_str = output.split("\n")[0].split(" ")[1]
        return self._parse_version(version_str)

    def command(self, cmd, print_greeting=True):
        s = TcpSocket(self._console_host, self._console_port)
        try:
            s.connect()
            greeting = s.read(128).decode()
            if print_greeting:
                self._logger.info(greeting)

            if isinstance(cmd, str):
                cmd = cmd.encode("utf-8")

            s.write(cmd + b"\n")

            data = s.read_until(b"...\n").decode()
            data = yaml.full_load(data)
            return data
        finally:
            s.close()


class TarantoolAsyncInstance(TarantoolInstance):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._is_stopping = False
        self._transport = None
        self._protocol = None
        self._last_return_code = None
        self._stop_event = asyncio.Event()

    @property
    def pid(self):
        return self._protocol.pid if self._protocol else None

    def prepare(self, recreate):
        self._last_return_code = None
        return super().prepare(recreate)

    def _on_process_exit(self, return_code):
        self._last_return_code = return_code
        if self._is_stopping:
            return
        self._stop_event.set()
        self.cleanup()

    async def wait_stopped(self):
        return await self._stop_event.wait()

    async def version(self):
        return await self.command("box.info.version")

    async def command(self, cmd, print_greeting=True):
        reader, writer = await asyncio.open_connection(
            self._console_host, self._console_port
        )

        greeting = (await reader.read(128)).decode()
        if print_greeting:
            self._logger.info(greeting)

        try:
            if isinstance(cmd, str):
                cmd = cmd.encode("utf-8")
            writer.write(cmd + b"\n")
            data = (await reader.readuntil(b"...\n")).decode()
            data = yaml.full_load(data)
            return data
        finally:
            writer.close()

    async def start(self, *, wait=True, recreate=True):
        self._logger.info("Starting Tarantool instance (%s)", self._title)
        self._stop_event.clear()
        initlua_path = self.prepare(recreate)
        self._logger.info("Launching process")

        factory = functools.partial(
            TarantoolInstanceProtocol, self, self._on_process_exit
        )
        if not self._command_args:
            args = [initlua_path]
        else:
            args = self._command_args
        (
            self._transport,
            self._protocol,
        ) = await asyncio.get_running_loop().subprocess_exec(
            factory,
            self._command_to_run,
            *args,
            stdin=None,
            stderr=asyncio.subprocess.PIPE,
        )

        if not wait:
            self._is_running = True
            return

        interval = 0.1
        attempts = math.ceil(self._timeout / interval)
        while attempts > 0:
            if self._protocol is None or self._protocol.returncode is not None:
                raise RuntimeError(
                    "{} exited unexpectedly with exit code {}".format(
                        self.fingerprint, self._last_return_code
                    )
                )
            try:
                status = await self.command("box.info.status", print_greeting=False)
                if status:
                    status = status[0]
                    if status == "running":
                        self._logger.info("Moved to the running state")
                        break
            except OSError:
                pass
            await asyncio.sleep(interval)
            attempts -= 1
        else:
            raise asyncio.TimeoutError(
                "Timeout while waiting for Tarantool to move to running state"
            )
        self._is_running = True

    async def stop(self):
        if self._protocol is not None:
            self._is_stopping = True
            self._protocol.terminate()

            if not self._is_running:
                return

            self._logger.info("Waiting for process to complete")
            await self._protocol.wait()
            self.cleanup()

    def terminate(self):
        if self._protocol is not None:
            self._is_stopping = True
            self._protocol.terminate()
            self.cleanup()

    def kill(self):
        if self._protocol is not None:
            self._is_stopping = True
            self._protocol.kill()
            self.cleanup()

    def cleanup(self):
        return_code = self._protocol.returncode
        self._logger.info("Finished with return code %d", return_code)

        self._is_stopping = False
        if self._transport:
            self._transport.close()
        self._transport = None
        self._protocol = None
        self._stop_event.clear()

        super().cleanup()


class TarantoolSyncDockerInstance(TarantoolSyncInstance):
    def __init__(
        self,
        *,
        docker_image=None,
        docker_tag=None,
        host="0.0.0.0",
        port=3301,
        console_host=None,
        console_port=3302,
        replication_source=None,
        title=None,
        logger=None,
        log_level=5,
        slab_alloc_arena=0.1,
        wal_mode="none",
        initlua_template=None,
        applua="-- app.lua --",
        timeout=10.0,
    ):
        super().__init__(
            host=host,
            port=port,
            console_host=console_host,
            console_port=console_port,
            replication_source=replication_source,
            title=title,
            logger=logger,
            log_level=log_level,
            slab_alloc_arena=slab_alloc_arena,
            wal_mode=wal_mode,
            root=None,
            specify_work_dir=False,
            cleanup=True,
            initlua_template=initlua_template,
            applua=applua,
            timeout=timeout,
        )
        self._docker_image = docker_image or "tarantool/tarantool"
        self._docker_tag = docker_tag or "1"

        cmd = (
            "docker run --rm "
            "-p {port}:{port} "
            "-p {console_port}:{console_port} "
            "-v {root}:/opt/tarantool "
            "{docker_image}:{docker_tag} "
            "tarantool /opt/tarantool/init.lua"
        )
        cmd = cmd.format(
            port=self.port,
            console_port=self.console_port,
            root=self._root,
            docker_image=self._docker_image,
            docker_tag=self._docker_tag,
        )
        self.logger.debug(cmd)
        args = cmd.split(" ")
        self._command_to_run = args[0]
        self._command_args = args[1:]

    @property
    def bin_version(self) -> Optional[tuple]:
        return self._parse_version(self._docker_tag)



import asyncio
import collections
from typing import Optional

cimport cpython
cimport cpython.dict
cimport cpython.list
cimport cython
from libc cimport stdio
from libc.stdint cimport uint32_t

from asynctnt.log import logger


@cython.final
cdef class IProtoErrorStackFrame:
    def __repr__(self):
        return "<Frame type={} code={} message={}>".format(
            self.error_type,
            self.code,
            self.message,
        )

@cython.final
cdef class IProtoError:
    pass

@cython.final
@cython.freelist(REQUEST_FREELIST)
cdef class Response:
    """
        Response object for all the requests to Tarantool
    """

    def __cinit__(self):
        self.request_ = None
        self.sync_ = 0
        self.code_ = -1
        self.return_code_ = -1
        self.schema_id_ = -1
        self.errmsg = None
        self.error = None
        self._rowcount = 0
        self.body = None
        self.encoding = None
        self.metadata = None
        self.params = None
        self.params_count = 0
        self.autoincrement_ids = None
        self.stmt_id_ = 0

        self._push_subscribe = False
        self._q = None
        self._push_event = None
        self._q_append = None
        self._q_popleft = None
        self._push_event_set = None
        self._push_event_clear = None

    cdef inline bint is_error(self):
        return self.code_ >= 0x8000

    # noinspection PyAttributeOutsideInit
    cdef inline void init_push(self):
        self._push_subscribe = True
        self._q = collections.deque()
        self._push_event = asyncio.Event()

        self._q_append = self._q.append
        self._q_popleft = self._q.popleft
        self._push_event_set = self._push_event.set
        self._push_event_clear = self._push_event.clear

    cdef inline void add_push(self, push):
        if not self._push_subscribe:
            return

        self._q_append(push)
        self.notify()

    cdef inline int push_len(self):
        return len(self._q)

    cdef inline object pop_push(self):
        if not self._push_subscribe:
            raise RuntimeError('Cannot pop push from a non-async response')

        push = self._q_popleft()
        if len(self._q) == 0:
            self._push_event_clear()
        return push

    cdef inline void set_data(self, list data):
        self.body = data
        self.notify()

    cdef inline void set_exception(self, exc):
        self._exception = exc
        self.notify()

    cdef inline object get_exception(self):
        return self._exception

    cdef inline void notify(self):
        if self._push_subscribe:
            self._push_event_set()  # Notify that there is no more data

    def __repr__(self):  # pragma: nocover
        data = self.body
        if data is not None:
            if len(data) > 10:
                parts = map(lambda x: ', '.join(map(repr, x)), [
                    data[:5],
                    data[-2:]
                ])

                data = ' ... '.join(parts)
                data = '[' + data + ']'

        return '<{} sync={} rowcount={} data={}>'.format(
            self.__class__.__name__, self.sync, self.rowcount, data)

    @property
    def sync(self) -> int:
        """
            Response's sync (incremental id) for the corresponding request
        """
        return <int> self.sync_

    @property
    def code(self) -> int:
        """
            Response code (0 - success)
        """
        return <int> self.code_

    @property
    def return_code(self) -> int:
        """
            Response return code (It's essentially a code & 0x7FFF)
        """
        return <int> self.return_code_

    @property
    def schema_id(self) -> int:
        return <int> self.schema_id_

    @property
    def rowcount(self) -> int:
        if self.body is not None:
            self_len = self._len()
            if self_len > 0:
                return self_len
        return self._rowcount

    @property
    def stmt_id(self) -> Optional[int]:
        return self.stmt_id_

    def done(self):
        return self.code_ >= 0

    cdef inline uint32_t _len(self):
        return <uint32_t> cpython.list.PyList_GET_SIZE(self.body)

    def __len__(self) -> int:
        if self.body is not None:
            return <int> self._len()
        return 0

    def __getitem__(self, i):
        return self.body[i]

    def __iter__(self):
        return iter(self.body)

cdef object _decode_obj(const char ** p, bytes encoding):
    cdef:
        uint32_t i
        mp_type obj_type

        const char *s
        uint32_t s_len

        uint32_t arr_size
        list arr

        uint32_t map_size
        dict map
        mp_type map_key_type
        const char *map_key_str
        uint32_t map_key_len
        object map_key

        int8_t ext_type
        IProtoDateTime dt

    obj_type = mp_typeof(p[0][0])
    if obj_type == MP_UINT:
        return mp_decode_uint(p)
    elif obj_type == MP_INT:
        return mp_decode_int(p)
    elif obj_type == MP_STR:
        s = NULL
        s_len = 0
        s = mp_decode_str(p, &s_len)
        try:
            return decode_string(s[:s_len], encoding)
        except UnicodeDecodeError:
            return <bytes> s[:s_len]
    elif obj_type == MP_BIN:
        s = NULL
        s_len = 0
        s = mp_decode_bin(p, &s_len)
        return <bytes> s[:s_len]
    elif obj_type == MP_BOOL:
        return mp_decode_bool(p)
    elif obj_type == MP_FLOAT:
        return mp_decode_float(p)
    elif obj_type == MP_DOUBLE:
        return mp_decode_double(p)
    elif obj_type == MP_ARRAY:
        arr_size = mp_decode_array(p)
        value = cpython.list.PyList_New(arr_size)
        for i in range(arr_size):
            el = _decode_obj(p, encoding)
            cpython.Py_INCREF(el)
            cpython.list.PyList_SET_ITEM(value, i, el)
        return value
    elif obj_type == MP_MAP:
        map = {}
        map_size = mp_decode_map(p)
        for i in range(map_size):
            map_key_type = mp_typeof(p[0][0])
            if map_key_type == MP_STR:
                map_key_len = 0
                map_key_str = mp_decode_str(p, &map_key_len)
                map_key = <object> (
                    decode_string(map_key_str[:map_key_len], encoding)
                )
            elif map_key_type == MP_UINT:
                map_key = <object> mp_decode_uint(p)
            elif map_key_type == MP_INT:
                map_key = <object> mp_decode_int(p)
            elif map_key_type == MP_FLOAT:
                map_key = <object> mp_decode_float(p)
            elif map_key_type == MP_DOUBLE:
                map_key = <object> mp_decode_double(p)
            else:  # pragma: nocover
                mp_next(p)  # skip current key
                mp_next(p)  # skip value
                logger.warning('Unexpected key type in map: %s',
                               map_key_type)
                continue

            map[map_key] = _decode_obj(p, encoding)

        return map
    elif obj_type == MP_NIL:
        mp_next(p)
        return None
    elif obj_type == MP_EXT:
        ext_type = 0
        s_len = mp_decode_extl(p, &ext_type)
        if ext_type == tarantool.MP_DECIMAL:
            return decimal_decode(p, s_len)
        elif ext_type == tarantool.MP_UUID:
            return uuid_decode(p, s_len)
        elif ext_type == tarantool.MP_ERROR:
            return parse_iproto_error(p, encoding)
        elif ext_type == tarantool.MP_DATETIME:
            datetime_zero(&dt)
            datetime_decode(p, s_len, &dt)
            return datetime_to_py(&dt)
        else:  # pragma: nocover
            logger.warning('Unexpected ext type: %d', ext_type)
            p += s_len  # skip unknown ext
            return None
    else:  # pragma: nocover
        mp_next(p)
        logger.warning('Unexpected obj type: %s', obj_type)
        return None

cdef list _response_parse_body_data(const char ** b,
                                    Response resp, BaseRequest req):
    cdef:
        uint32_t size
        uint32_t tuple_size
        list tuples
        uint32_t i

        Metadata metadata

    size = mp_decode_array(b)
    tuples = []

    if req.parse_as_tuples:
        # decode as TarantoolTuples

        metadata = resp.metadata
        if metadata is None:
            metadata = req.metadata()

        for i in range(size):
            if mp_typeof(b[0][0]) != MP_ARRAY:  # pragma: nocover
                raise TypeError(
                    'Tuple must be an array when decoding as TarantoolTuple'
                )

            tuple_size = mp_decode_array(b)
            t = tupleobj.AtntTuple_New(metadata, <int> tuple_size)
            for i in range(tuple_size):
                value = _decode_obj(b, resp.encoding)
                cpython.Py_INCREF(value)
                tupleobj.AtntTuple_SET_ITEM(t, i, value)

            tuples.append(t)
    else:
        # decode as raw objects
        for i in range(size):
            tuples.append(_decode_obj(b, resp.encoding))

    return tuples

cdef ssize_t response_parse_header(const char *buf, uint32_t buf_len,
                                   Header *hdr) except -1:
    cdef:
        const char *b
        uint32_t size
        uint32_t key

    b = <const char *> buf
    # mp_fprint(stdio.stdout, b)
    # stdio.fprintf(stdio.stdout, "\n")

    if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
        raise TypeError('Response header must be a MP_MAP')

    # parsing header
    size = mp_decode_map(&b)
    for _ in range(size):
        if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
            raise TypeError('Header key must be a MP_UINT')

        key = mp_decode_uint(&b)
        if key == tarantool.IPROTO_REQUEST_TYPE:
            if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
                raise TypeError('code type must be a MP_UINT')

            hdr.code = <uint32_t> mp_decode_uint(&b)
            hdr.return_code = hdr.code & 0x7FFF
        elif key == tarantool.IPROTO_SYNC:
            if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
                raise TypeError('sync type must be a MP_UINT')

            hdr.sync = mp_decode_uint(&b)
        elif key == tarantool.IPROTO_SCHEMA_VERSION:
            if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
                raise TypeError('schema_id type must be a MP_UINT')

            hdr.schema_id = mp_decode_uint(&b)
        else:  # pragma: nocover
            logger.warning(
                'Unknown key with code \'%d\' in header. Skipping.', key)
            mp_next(&b)

    return <ssize_t> (b - buf)

cdef Metadata response_parse_metadata(const char ** b, bytes encoding):
    cdef:
        uint32_t arr_size
        uint32_t field_map_size
        uint32_t key
        uint32_t s_len
        uint32_t i
        const char *s
        Field field
        Metadata metadata

    metadata = <Metadata> Metadata.__new__(Metadata)
    arr_size = mp_decode_array(b)
    for i in range(arr_size):
        field_map_size = mp_decode_map(b)
        if field_map_size == 0:  # pragma: nocover
            raise RuntimeError('Field map must contain at least '
                               '1 element - field_name')

        field = <Field> Field.__new__(Field)
        field_id = i
        for _ in range(field_map_size):
            key = mp_decode_uint(b)
            if key == tarantool.IPROTO_FIELD_NAME:
                s = NULL
                s_len = 0
                s = mp_decode_str(b, &s_len)
                field.name = \
                    decode_string(s[:s_len], encoding)

            elif key == tarantool.IPROTO_FIELD_TYPE:
                s = NULL
                s_len = 0
                s = mp_decode_str(b, &s_len)
                field.type = \
                    decode_string(s[:s_len], encoding)

            elif key == tarantool.IPROTO_FIELD_COLL:
                s = NULL
                s_len = 0
                s = mp_decode_str(b, &s_len)
                field.collation = \
                    decode_string(s[:s_len], encoding)

            elif key == tarantool.IPROTO_FIELD_IS_NULLABLE:
                field.is_nullable = mp_decode_bool(b)

            elif key == tarantool.IPROTO_FIELD_IS_AUTOINCREMENT:
                field.is_autoincrement = mp_decode_bool(b)

            elif key == tarantool.IPROTO_FIELD_SPAN:
                if mp_typeof(b[0][0]) == MP_NIL:  # pragma: nocover
                    mp_next(b)
                    field.span = None

                elif mp_typeof(b[0][0]) == MP_STR:
                    s = NULL
                    s_len = 0
                    s = mp_decode_str(b, &s_len)
                    field.span = \
                        decode_string(s[:s_len], encoding)

                else:  # pragma: nocover
                    raise TypeError(
                        "IPROTO_FIELD_SPAN must be either STR or NIL"
                    )
            else:  # pragma: nocover
                logger.debug(
                    'unknown key in metadata decoding: %d', key)
                mp_next(b)

        if field.name is None:  # pragma: nocover
            raise RuntimeError('field.name must not be None')

        metadata.add(<int> field_id, field)
    return metadata

cdef inline IProtoErrorStackFrame parse_iproto_error_stack_frame(const char ** b, bytes encoding):
    cdef:
        uint32_t size
        uint32_t key
        const char * s
        uint32_t s_len
        IProtoErrorStackFrame frame
        uint32_t unum

    size = 0
    key = 0

    frame = <IProtoErrorStackFrame> IProtoErrorStackFrame.__new__(IProtoErrorStackFrame)

    size = mp_decode_map(b)
    for _ in range(size):
        key = mp_decode_uint(b)

        if key == tarantool.MP_ERROR_TYPE:
            s = NULL
            s_len = 0
            s = mp_decode_str(b, &s_len)
            frame.error_type = decode_string(s[:s_len], encoding)

        elif key == tarantool.MP_ERROR_FILE:
            s = NULL
            s_len = 0
            s = mp_decode_str(b, &s_len)
            frame.file = decode_string(s[:s_len], encoding)

        elif key == tarantool.MP_ERROR_LINE:
            frame.line = <int> mp_decode_uint(b)

        elif key == tarantool.MP_ERROR_MESSAGE:
            s = NULL
            s_len = 0
            s = mp_decode_str(b, &s_len)
            frame.message = decode_string(s[:s_len], encoding)

        elif key == tarantool.MP_ERROR_ERRNO:
            frame.err_no = <int> mp_decode_uint(b)

        elif key == tarantool.MP_ERROR_ERRCODE:
            frame.code = <int> mp_decode_uint(b)

        elif key == tarantool.MP_ERROR_FIELDS:
            if mp_typeof(b[0][0]) != MP_MAP:  # pragma: nocover
                raise TypeError(f'iproto_error stack frame fields must be a '
                                f'map, but got {mp_typeof(b[0][0])}')

            frame.fields = _decode_obj(b, encoding)

        else:  # pragma: nocover
            logger.debug(f"unknown iproto_error stack element with key {key}")
            mp_next(b)

    return frame

cdef inline IProtoError parse_iproto_error(const char ** b, bytes encoding):
    cdef:
        uint32_t size
        uint32_t arr_size
        uint32_t key
        uint32_t i
        IProtoError error

    size = 0
    arr_size = 0
    key = 0

    error = <IProtoError> IProtoError.__new__(IProtoError)

    size = mp_decode_map(b)
    for _ in range(size):
        key = mp_decode_uint(b)

        if key == tarantool.MP_ERROR_STACK:
            arr_size = mp_decode_array(b)
            error.trace = cpython.list.PyList_New(arr_size)
            for i in range(arr_size):
                el = parse_iproto_error_stack_frame(b, encoding)
                cpython.Py_INCREF(el)
                cpython.list.PyList_SET_ITEM(error.trace, i, el)
        else:  # pragma: nocover
            logger.debug(f"unknown iproto_error map field with key {key}")
            mp_next(b)

    return error

cdef ssize_t response_parse_body(const char *buf, uint32_t buf_len,
                                 Response resp, BaseRequest req,
                                 bint is_chunk) except -1:
    cdef:
        const char *b
        uint32_t size
        uint32_t arr_size
        uint32_t field_map_size
        uint32_t key
        uint32_t s_len
        uint32_t i
        const char *s
        list data
        Field field

    b = <const char *> buf
    # mp_fprint(stdio.stdout, b)
    # stdio.fprintf(stdio.stdout, "\n")

    # parsing body

    if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
        raise TypeError('Response body must be a MP_MAP')

    size = mp_decode_map(&b)
    for _ in range(size):
        if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
            raise TypeError('Header key must be a MP_UINT')

        key = mp_decode_uint(&b)
        if key == tarantool.IPROTO_ERROR_24:
            if mp_typeof(b[0]) != MP_STR:  # pragma: nocover
                raise TypeError('errstr type must be a MP_STR')

            s = NULL
            s_len = 0
            s = mp_decode_str(&b, &s_len)
            resp.errmsg = decode_string(s[:s_len], resp.encoding)

        elif key == tarantool.IPROTO_ERROR:
            if mp_typeof(b[0]) != MP_MAP:  # pragma: nocover
                raise TypeError('IPROTO_ERROR type must be a MP_MAP')

            resp.error = parse_iproto_error(&b, resp.encoding)

        elif key == tarantool.IPROTO_STMT_ID:
            if mp_typeof(b[0]) != MP_UINT:  # pragma: nocover
                raise TypeError(f'IPROTO_STMT_ID type must be a MP_UINT, but got {mp_typeof(b[0])}')
            resp.stmt_id_ = mp_decode_uint(&b)

        elif key == tarantool.IPROTO_METADATA:
            if not req.parse_metadata:
                mp_next(&b)
                continue

            resp.metadata = response_parse_metadata(&b, resp.encoding)

        elif key == tarantool.IPROTO_BIND_METADATA:
            if not req.parse_metadata:
                mp_next(&b)
                continue

            resp.params = response_parse_metadata(&b, resp.encoding)

        elif key == tarantool.IPROTO_BIND_COUNT:

            resp.params_count = <int> mp_decode_uint(&b)

        elif key == tarantool.IPROTO_SQL_INFO:
            field_map_size = mp_decode_map(&b)
            if field_map_size == 0:
                raise RuntimeError('Field map must contain at least '
                                   '1 element - rowcount')

            for _ in range(field_map_size):
                key = mp_decode_uint(&b)
                if key == tarantool.SQL_INFO_ROW_COUNT:
                    resp._rowcount = mp_decode_uint(&b)
                elif key == tarantool.SQL_INFO_AUTOINCREMENT_IDS:
                    arr_size = mp_decode_array(&b)
                    ids = cpython.list.PyList_New(arr_size)
                    for i in range(arr_size):
                        el = <object> mp_decode_uint(&b)
                        cpython.Py_INCREF(el)
                        cpython.list.PyList_SET_ITEM(ids, i, el)
                    resp.autoincrement_ids = ids
                else:
                    logger.debug('unknown key in sql info decoding: %d', key)
                    mp_next(&b)

        elif key == tarantool.IPROTO_DATA:
            if mp_typeof(b[0]) != MP_ARRAY:  # pragma: nocover
                raise TypeError('body data type must be a MP_ARRAY')
            data = _response_parse_body_data(&b, resp, req)
            if is_chunk:
                resp.add_push(data)
            else:
                resp.set_data(data)

        elif key == tarantool.IPROTO_VERSION:
            logger.debug("IProto version: %s", _decode_obj(&b, resp.encoding))

        elif key == tarantool.IPROTO_FEATURES:
            logger.debug("IProto features available: %s", _decode_obj(&b, resp.encoding))

        elif key == tarantool.IPROTO_AUTH_TYPE:
            logger.debug("IProto auth type: %s", _decode_obj(&b, resp.encoding))

        else:  # pragma: nocover
            logger.debug('unknown key in body map: %s', hex(int(key)))
            mp_next(&b)

    return <ssize_t> (b - buf)



import datetime
import sys
import uuid
from dataclasses import dataclass
from decimal import Decimal

import dateutil.parser
import pytz

from asynctnt import IProtoError
from asynctnt.exceptions import ErrorCode, TarantoolDatabaseError
from tests import BaseTarantoolTestCase
from tests._testbase import ensure_version


@dataclass
class DecimalTestCase:
    python: Decimal
    tarantool: str


class MpExtTestCase(BaseTarantoolTestCase):
    @ensure_version(min=(2, 2))
    async def test__decimal(self):
        space = "tester_ext_dec"

        cases = [
            "-12.34",
            "-12.345",
            "-12.4",
            "0.000",
            "42",
            "0.33",
            "0.000000000000000000000000000000000010",
            "-0.000000000000000000000000000000000010",
            "0.1111111111111111",
            "-0.1111111111111111",
            "-0.111111",
            "-18.34",
            "-108.123456789",
            "100",
            "0.1",
            "-0.1",
            "2.718281828459045",
            "-2.718281828459045",
            "3.141592653589793",
            "-3.141592653589793",
            "1",
            "-1",
            "0",
            "-0",
            "0.01",
            "-0.01",
            "0.001",
            "11111111111111111111111111111111111111",
            "-11111111111111111111111111111111111111",
            "0.0000000000000000000000000000000000001",
            "-0.0000000000000000000000000000000000001",
            "0.00000000000000000000000000000000000009",
            "-0.00000000000000000000000000000000000009",
            "99999999999999999999999999999999999999",
            "-99999999999999999999999999999999999999",
            "1234567891234567890.0987654321987654321",
            "-1234567891234567890.0987654321987654321",
            "1e33",
            "1.2345e33",
            "1.2345e2",
            "1.2345e4",
            "-1e33",
            "1e-33",
        ]

        for case in cases:
            with self.subTest(case):
                dec = Decimal(case)
                res = await self.conn.replace(space, [1, dec])
                self.assertEqual(res[0][1], dec, "self-return works")

                res = await self.conn.eval(
                    f"local decimal = require('decimal'); return decimal.new('{case}')"
                )
                self.assertEqual(res[0], dec, "matches tarantool decimal")

    @ensure_version(min=(2, 4, 1))
    async def test__uuid(self):
        space = "tester_ext_uuid"

        val = uuid.uuid4()
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.UUID("f6423bdf-b49e-4913-b361-0740c9702e4b")
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.UUID("00000000-0000-0000-0000-000000000000")
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.uuid1(1, 100)
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.uuid3(uuid.uuid4(), "hellothere")
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.uuid5(uuid.NAMESPACE_URL, "generalkenobi")
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

    @ensure_version(min=(2, 4, 1))
    async def test__ext_error(self):
        try:
            await self.conn.eval(
                """
                box.schema.space.create('_space')
            """
            )
        except TarantoolDatabaseError as e:
            self.assertIsNotNone(e.error)
            self.assertGreater(len(e.error.trace), 0)
            frame = e.error.trace[0]
            self.assertEqual("ClientError", frame.error_type)
            self.assertIsNotNone(frame.file)
            self.assertIsNotNone(frame.line)
            self.assertEqual("Space '_space' already exists", frame.message)
            self.assertEqual(0, frame.err_no)
            self.assertEqual(ErrorCode.ER_SPACE_EXISTS, frame.code)

    @ensure_version(min=(2, 4, 1))
    async def test__ext_error_custom(self):
        try:
            await self.conn.eval(
                """
                local e = box.error.new{code=5,reason='A',type='B'}
                box.error(e)
            """
            )
        except TarantoolDatabaseError as e:
            self.assertIsNotNone(e.error)
            self.assertGreater(len(e.error.trace), 0)
            frame = e.error.trace[0]
            self.assertEqual("CustomError", frame.error_type)
            self.assertIsNotNone(frame.file)
            self.assertIsNotNone(frame.line)
            self.assertEqual("A", frame.message)
            self.assertEqual(0, frame.err_no)
            self.assertEqual(5, frame.code)
            self.assertIn("custom_type", frame.fields)
            self.assertEqual("B", frame.fields["custom_type"])

    @ensure_version(min=(2, 10))
    async def test__ext_error_custom_return(self):
        resp = await self.conn.eval(
            """
            local e = box.error.new{code=5,reason='A',type='B'}
            return e
        """
        )
        e = resp[0]
        self.assertIsInstance(e, IProtoError)
        self.assertGreater(len(e.trace), 0)
        frame = e.trace[0]
        self.assertEqual("CustomError", frame.error_type)
        self.assertEqual("eval", frame.file)
        self.assertEqual(2, frame.line)
        self.assertEqual("A", frame.message)
        self.assertEqual(0, frame.err_no)
        self.assertEqual(5, frame.code)
        self.assertIn("custom_type", frame.fields)
        self.assertEqual("B", frame.fields["custom_type"])

    @ensure_version(min=(2, 10))
    async def test__ext_error_custom_return_with_disabled_exterror(self):
        await self.conn.eval(
            """
            require('msgpack').cfg{encode_error_as_ext = false}
        """
        )
        try:
            resp = await self.conn.eval(
                """
                local e = box.error.new{code=5,reason='A',type='B'}
                return e
            """
            )
            e = resp[0]
            self.assertIsInstance(e, str)
            self.assertEqual("A", e)
        finally:
            await self.conn.eval(
                """
                require('msgpack').cfg{encode_error_as_ext = true}
            """
            )

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_read(self):
        resp = await self.conn.eval(
            """
            local date = require('datetime')
            return date.parse('2000-01-01T02:00:00.23+0300')
        """
        )
        res = resp[0]
        dt = datetime_fromisoformat("2000-01-01T02:00:00.230000+03:00")
        self.assertEqual(dt, res)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_tz(self):
        resp = await self.conn.eval(
            """
            local date = require('datetime')
            return date.parse('2000-01-01T02:00:00 MSK')
        """
        )
        res = resp[0]
        dt = datetime_fromisoformat("2000-01-01T02:00:00+03:00")
        self.assertEqual(dt, res)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_read_neg_tz(self):
        resp = await self.conn.eval(
            """
            local date = require('datetime')
            return date.parse('2000-01-01T02:17:43.23-08:00')
        """
        )
        res = resp[0]
        dt = datetime_fromisoformat("2000-01-01T02:17:43.230000-08:00")
        self.assertEqual(dt, res)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_read_before_1970(self):
        resp = await self.conn.eval(
            """
            local date = require('datetime')
            return date.parse('1930-01-01T02:17:43.23-08:00')
        """
        )
        res = resp[0]
        dt = datetime_fromisoformat("1930-01-01T02:17:43.230000-08:00")
        self.assertEqual(dt, res)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write(self):
        sp = "tester_ext_datetime"
        dt = datetime_fromisoformat("2000-01-01T02:17:43.230000-08:00")
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res["dt"])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_before_1970(self):
        sp = "tester_ext_datetime"
        dt = datetime_fromisoformat("1004-01-01T02:17:43.230000+04:00")
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res["dt"])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_without_tz(self):
        sp = "tester_ext_datetime"
        dt = datetime_fromisoformat("2022-04-23T02:17:43.450000")
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res["dt"])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_without_tz_integer(self):
        sp = "tester_ext_datetime"
        dt = datetime_fromisoformat("2022-04-23T02:17:43")
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res["dt"])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_pytz(self):
        sp = "tester_ext_datetime"
        dt = datetime_fromisoformat("2022-04-23T02:17:43")
        dt = pytz.timezone("Europe/Amsterdam").localize(dt)
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res["dt"])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_pytz_america(self):
        sp = "tester_ext_datetime"
        dt = datetime_fromisoformat("2022-04-23T02:17:43")
        dt = pytz.timezone("America/New_York").localize(dt)
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res["dt"])


def datetime_fromisoformat(s):
    if sys.version_info < (3, 7, 0):
        return dateutil.parser.isoparse(s)
    return datetime.datetime.fromisoformat(s)
